#
# Sensus Consulting Ltd, Copyright 1997-1998
#
# Matt Newman <matt@sensus.org>
#
#INDEX\
proc itk::Archetype {} {}
#
# Derived from:
#
# itk::Archetype
# ----------------------------------------------------------------------
# Base class for all widgets in the [incr Tk] Toolkit.  Provides
# facilities to merge widget options into a composite list of options
# for the overall widget.  Derived classes add widgets and methods to
# specialize behavior.
#
#   METHODS:
#     configure
#     configure -option
#     configure -option value ?-option value?...
#        Used to set/query configuration options
#
#     component
#     component <name> <command> ?<arg> <arg>...?
#        Invokes the given <command> as a method on the component
#        called <name>.
#
#     itk_component add <name> <create-cmd> <option-cmds>
#        Creates a component widget and merges its options into
#        the composite option list for the overall widget
#
#     itk_component delete <name> ?<name>...?
#        Destroys a component widget and removes its options from
#        the composite option list
#
#     itk_option add <name> ?<name>...?
#        Adds the option <name> belonging to a class or component
#        widget into the option list.  Options can be added even
#        if they were not originally kept when the component was
#        created.
#
#     itk_option remove <name> ?<name>...?
#        Removes the option <name> belonging to a class or component
#        widget from the option list.  This allows a derived class
#        to turn off or redefine undesirable options inherited from
#        a base class.
#
#   WIDGET ATTRIBUTES:
#     none
#
# ----------------------------------------------------------------------
#   AUTHOR:  Michael J. McLennan
#            Bell Labs Innovations for Lucent Technologies
#            mmclennan@lucent.com
#            http://www.tcltk.com/itcl
#
#      RCS:  $Id$
# ----------------------------------------------------------------------
#               Copyright (c) 1993-1996  Lucent Technologies
# ======================================================================
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies and that
# both that the copyright notice and warranty disclaimer appear in
# supporting documentation, and that the names of Lucent Technologies
# any of their entities not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# Lucent Technologies disclaims all warranties with regard to this
# software, including all implied warranties of merchantability and
# fitness.  In no event shall Lucent be liable for any special, indirect
# or consequential damages or any damages whatsoever resulting from loss
# of use, data or profits, whether in an action of contract, negligence
# or other tortuous action, arising out of or in connection with the use
# or performance of this software.
# ======================================================================

class itk::Archetype {
	public {
		constructor {args} {}
		destructor {}

		method cget {option}
		method configure {args}
		method component {args}

		itk_option define -clientdata clientData ClientData ""
	}
	protected {
		method itk_initialize {args}
		method itk_option {option args}
		method itk_component {option name constructor {keep ""}}
		method itk_dispatch {class option}
		method bindtags {w args}
		# For timing tests
		method nop {args} {}

		variable itk_option
		variable itk_component
		variable itk_default
		variable itk_interior
		variable itk_keep
		variable itk_path
	}
	private method __addOption {class option resName resClass defVal}
}

body itk::Archetype::constructor {args} {
	set w [namespace tail $this]
	if {![string match .* $w]} {
		error "bad window path name \"$w\""
	}
	eval itk_initialize $args
}
body itk::Archetype::destructor {} {
}
body itk::Archetype::cget {option} {
	if {![_info exists itk_option($option)]} {
		error "unknown option \"$option\""
	}
	return $itk_option($option)
}
body itk::Archetype::configure {args} {
	switch -- [llength $args] {
	0	{
		foreach opt [lsort [array names itk_option -*]] {
			set rec $itk_default($opt)
			lappend rec $itk_option($opt)
			set conf($opt) $rec
		}
		set ret {}
		foreach opt [lsort [array names conf]] {
			lappend ret $conf($opt)
		}
		return $ret
	}
	1	{
		set opt [lindex $args 0]
		if {![_info exists itk_option($opt)]} {
			error "unknown option \"$opt\""
		}
		set conf $itk_default($opt)
		lappend conf $itk_option($opt)
		return $conf
	}
	default	{
		set class [info class]
		ainit work
		foreach {opt val} $args {
			if {![_info exists itk_option($opt)]} {
				error "unknown option \"$opt\""
			}
			set itk_option($opt) $val
			if {[_info exists ${class}::_info(variable,[string range $opt 1 end])]} {
				set __v__([string range $opt 1 end]) $val
			}
			itk_dispatch $class [string range $opt 1 end]
			#
			# Bunch up keeps to minimize widget
			# reconfigure overhead - which can be
			# quite high during the construction phase.
			#
			set items [array get itk_keep $opt,*]
			set slen [string length $opt,]
			foreach {comp ropt} $items {
				set comp [string range $comp $slen end]
				if {$comp == "hull"} {
					set path $itk_path
				} else {
					set path $itk_component($comp)
				}
				lappend work($path) $ropt $itk_option($opt)
			}
		}
		#
		# OK, now dispatch keeps
		#
		foreach {path config} [array get work] {
			#tclLog ">>> $path configure $work($path)"
			eval $path configure $config
		}
		return ""
	}
	};#sw
}
#
# Dispatch config bodies
#
body itk::Archetype::itk_dispatch {class option} {
	upvar ${class}::_info info

	#tclLog "$this itk_dispatch $class $option"
	# Dispatch super class config's first.
	foreach super [lrange $info(heritage) 1 end] {
		itk_dispatch $super $option
	}
	# Run config script in correct scope.
	if {[_info exists info(config,$option)] && $info(config,$option) != ""} {
		${class}::_eval $info(config,$option)
	}
}
body itk::Archetype::component {args} {
	set argc [llength $args]
	switch -- $argc {
	0	{return [array names itk_component]}
	1	{return $itk_component([lindex $args 0])}
	default {
		set name [lindex $args 0]
		set args [lrange $args 1 end]
		if {$name == "hull"} {
			return [uplevel #0 $itk_path $args]
		} else {
			return [uplevel #0 $itk_component($name) $args]
		}
	}
	};#switch
}
body itk::Archetype::itk_initialize {args} {
	set class [uplevel 1 namespace current]
	upvar ${class}::_info info

	#tclLog "${class}: itk_initialize $args"

	foreach {option data} [array get info variable,*] {
		set option [string range $option 9 end]

		#tclLog "($class)$this: itk_initialize $data"
		if {[lindex $data 0] != "public"} continue
		# public variable, fake up as option
		if [_info exists __v__($option)] {
			set rec [list $class -$option {} {} $__v__($option)]
		} else {
			set rec [list $class -$option {} {} {}]
		}
		set opts($option) $rec
	}
	foreach opt [array names info option,*] {
		set opt [string range $opt 7 end]
		set opts($opt) [lrange $info(option,$opt) 1 5]
	}
	foreach opt [array names opts] {
		#tclLog "eval __addOption $opts($opt)"
		eval __addOption $opts($opt)
	}
	if {$class == [info class]} {
		#
		# Most virtual class - fire off config scripts.
		# But only once mind you!
		#
		foreach {opt val} [array get itk_option] {
			set conf($opt) $val
		}
		foreach {opt val} $args {
			set conf($opt) $val
		}
		eval configure [array get conf]
	} else {
		if {[llength $args]>0} {
			eval configure $args
		}
	}
}
body itk::Archetype::itk_component {option args} {
	set class [uplevel 1 namespace current]

#	tclLog "($class) $this itk_component $option $args"
	switch -- $option {
	add	{
		if {[llength $args]!=2 && [llength $args]!=3} {
			error "wrong # args: should be \"itk_component add name create ?options?\""
		}
		set name [lindex $args 0]
		set create [lindex $args 1]
		set options [lindex $args 2]

		#
		# Create widget
		#
		if {[string compare $name hull]==0} {
			::rename $this ${this}_tmp_

			set itk_component($name)	[namespace tail [uplevel 1 $create]]
			set itk_hull [namespace tail $this]
			set itk_path [info class]::${itk_hull}_real_

			::rename $itk_hull $itk_path
			::rename ${this}_tmp_ $this
			set path $itk_path
		} else {
			set itk_component($name)	[namespace tail [uplevel 1 $create]]
			set path $itk_component($name)
			bindtags $path insert $path itk-destroy-$path
			bind itk-destroy-$path <Destroy> [code $this itk_component delete $name]

			if {[llength $args]==2} {
				set options [usual [winfo class $path]]
			}
		}
		#
		# Parse keeps, renames etc.
		#
		tk++::parser::parse [info class] $options
		upvar tk++::parser::__ __
		#
		# Process any options discovered
		#
		foreach ropt [array names __ -*] {
			set opt [lindex $__($ropt) 0]
			set conf [$path configure $ropt]
			lassign $conf ropt resN resC defV
			if {$ropt != $opt} {
				lassign $__($ropt) opt resN resC 
			}
			#tclLog "keep($name) [list __addOption $class $opt $resN $resC $defV]"
			__addOption $class $opt $resN $resC $defV

			set itk_keep($opt,$name) $ropt
		}
		return $name
	}
	delete	{
		foreach name $args {
			if {![_info exists itk_component($name)]} continue

			foreach key [array names itk_keep *,$name] {
				set opt [lindex [split $key ,] 1]
				itk_option remove $name.$opt
			}
			unset itk_component($name)
		}
	}
	default	{
		error "bad option \"$option\": should be one of add, or delete"
	}
	};#switch
}
body itk::Archetype::itk_option {option args} {
	set class [uplevel 1 namespace current]
	switch -- $option {
	add	{
		foreach opt $args {
			set tmp [split $opt .]
			set name [lindex $tmp 0]
			set opt -[lindex $tmp 1]

			if {$name == "hull"} {
				set path $itk_path
			} else {
				set path $itk_component($name)
			}
			eval __addOption $class [lrange [$path configure $opt] 0 3]

			set itk_keep($opt,$name) $opt
		}
	}
	remove	{
		foreach opt $args {
			if {[string match *.* $opt]} {
				set tmp [split $opt .]
				set name [lindex $tmp 0]
				set opt [lindex $tmp 1]

				if {![_info exists itk_component($name)]} {
					error "unknown component \"$name\""
				}
				# If not keep'd ignore
				if {![_info exists itk_keep($opt,$name)]} continue
				unset itk_keep($opt,$name)
				if {[array names itk_keep $opt,*]!=""} continue
				# If last keep, check to see if it was auto-created or not.
				if {![_info exists ${class}::_info(option,$opt)]} {
					unset itk_option(-$opt)
					unset itk_default(-$opt)
				}
			} elseif {[string match *::* $opt]} {
				set cls [namespace qualifiers $opt]
				set opt [namespace tail $opt]

				if {![_info exists ${cls}::_info(option,$opt)]} {
					error "option \"$opt\" not defined in class \"$cls\""
				}
				# If keeps exist ignore this request
				if {[array names itk_keep $opt,*]!=""} continue
				unset itk_option(-$opt)
				unset itk_default(-$opt)
			} else {
				error "bad option \"$opt\": must be one of class::option, or component.option"
			}
		}
	}
	default	{
		error "bad option \"$option\": must be one of add, or remove"
	}
	};#switch
}
body itk::Archetype::__addOption {class opt resName resClass defVal} {
	if {![_info exists itk_default($opt)]} {
		if {[_info exists itk_component(hull)]} {
			set dbVal [::option get $itk_component(hull) $resName $resClass]
			if {$dbVal != ""} {
				#tclLog "${class}: using option DB value for $opt"
				set defVal $dbVal
			}
		}
		set itk_default($opt) [list $opt $resName $resClass $defVal]
		set itk_option($opt) $defVal
	} else {
		# Check resource names match
	}
}
body itk::Archetype::bindtags {w args} {
	set argc [llength $args]
	if {$argc==0} {
		return [::bindtags $w]
	} elseif {$argc==1} {
		return [::bindtags $w [lindex $args 0]]
	}
	set bt	[::bindtags $w]
	switch -- [lindex $args 0] {
	insert {
			if {$argc < 3} {
				error "wrong # args: should be \"bindtags window insert index element ?element ...?\""
			}
			set idx [lindex $args 1]
			if {$idx != "end" && [string match {[0-9]*} $idx]==0} {
				set idx [lsearch $bt $idx]
			}
			::bindtags $w [eval linsert [list $bt $idx] [lrange $args 2 end]]
		}
	replace	{
			if {$argc < 3} {
				error "wrong # args: should be \"bindtags window replace first last ?element element ...?\""
			}
			set first [lindex $args 1]
			set last [lindex $args 2]
			if {[string match {[0-9]*} $first]==0} {
				set first [lsearch $bt $first]
			}
			if {$last != "end" && [string match {[0-9]*} $last]==0} {
				set last [lsearch $bt $last]
			}
			::bindtags $w [eval lreplace [list $bt $first $last] [lrange $args 3 end]]
		}
	default	{
			error "bad option \"[lindex $args 0]\": must be one of insert, or replace"
		}
	}
}
